# -*- coding: utf-8 -*-

# Copyright (c) 2009, 2010 Jack Kaliko <efrim@azylum.org> {{{
# Copyright (c) 2008 Rick van Hattem
#
#  This file is part of MPD_sima
#
#  MPD_sima is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  
#  MPD_sima is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  
#  You should have received a copy of the GNU General Public License
#  along with MPD_sima.  If not, see <http://www.gnu.org/licenses/>. 
#
#
#  }}}

#    DOC:#{{{
#    MuscicBrainz ID: <http://musicbrainz.org/doc/MusicBrainzIdentifier>
#    Artists: <http://musicbrainz.org/doc/Artist_Name>
#             <http://musicbrainz.org/doc/Same_Artist_With_Different_Names>
#}}}

import sqlite3

class SimaDB(object):
    "SQLite managment"

    def __init__(self, db_path=None):#{{{
        self._db_path = db_path
        self.create_db()

    def close_database_connection(self, connection):
        """Close the database connection."""
        connection.close()

    def get_database_connection(self):
        """get database reference"""
        connection = sqlite3.connect(
            self._db_path, timeout=5.0, isolation_level="immediate")
        #connection.text_factory = str
        return connection#}}}

    def get_artist(self, artist_name, mbid=None, with_connection=None, add_not=False):#{{{
        """get artist information from the database.
        if not in database insert new entry."""
        if with_connection:
            connection = with_connection
        else:
            connection = self.get_database_connection()
        rows = connection.execute(
            "SELECT * FROM artists WHERE name = ?", (artist_name,))
        for row in rows:
            if not with_connection:
                self.close_database_connection(connection)
            return row
        if add_not:
            return False
        connection.execute(
            "INSERT INTO artists (name, mbid) VALUES (?, ?)", (artist_name, mbid))
        connection.commit()
        rows = connection.execute(
            "SELECT * FROM artists WHERE name = ?", (artist_name,))
        for row in rows:
            if not with_connection:
                self.close_database_connection(connection)
            return row
        if not with_connection:
            self.close_database_connection(connection)#}}}

    def get_artists(self, with_connection=None):#{{{
        """Returns all artists in DB"""
        if with_connection:
            connection = with_connection
        else:
            connection = self.get_database_connection()
        rows = connection.execute("SELECT name FROM artists ORDER BY name")
        results = [row for row in rows]
        if not with_connection:
            self.close_database_connection(connection)
        for artist in results:
            yield artist#}}}

    def _set_mbid(self, artist_id=None, mbid=None, with_connection=None):#{{{
        """get artist information from the database.
        if not in database insert new entry."""
        if with_connection:
            connection = with_connection
        else:
            connection = self.get_database_connection()
        connection.execute("UPDATE artists SET mbid = ? WHERE id = ?",
            (mbid, artist_id))
        connection.commit()
        if not with_connection:
            self.close_database_connection(connection)#}}}

    def _get_similar_artists_from_db(self, artist_id):#{{{
        connection = self.get_database_connection()
        results = [row for row in connection.execute(
            "SELECT match, name FROM usr_artist_2_artist INNER JOIN"
            " artists ON usr_artist_2_artist.artist2 = artists.id WHERE"
            " usr_artist_2_artist.artist1 = ? ORDER BY match DESC;",
            (artist_id,))]
        self.close_database_connection(connection)
        for score, artist in results:
            yield {'score': score, 'artist': artist}#}}}

    def _get_reverse_similar_artists_from_db(self, artist_id):#{{{
        connection = self.get_database_connection()
        results = [row for row in connection.execute(
            "SELECT name FROM usr_artist_2_artist INNER JOIN"
            " artists ON usr_artist_2_artist.artist1 = artists.id WHERE"
            " usr_artist_2_artist.artist2 = ?;",
            (artist_id,))]
        self.close_database_connection(connection)
        for artist in results:
            yield artist[0]#}}}

    def get_similar_artists(self, track):#{{{
        """get similar artists from the database sorted by descending
        match score"""
        # TODO: remove Track object here and use Artist name only
        # TODO: raise exception if given artist name is not unicode
        artist_name = track.get_artist()
        artist_id = self.get_artist(artist_name)[0]
        for result in self._get_similar_artists_from_db(artist_id):
            yield result #}}}

    def _get_artist_match(self, artist1, artist2, with_connection=None):#{{{
        """get artist match score from database"""
        if with_connection:
            connection = with_connection
        else:
            connection = self.get_database_connection()
        rows = connection.execute(
            "SELECT match FROM usr_artist_2_artist WHERE artist1 = ?"
            " AND artist2 = ?",
            (artist1, artist2))
        result = 0
        for row in rows:
            result = row[0]
            break
        if not with_connection:
            self.close_database_connection(connection)
        return result#}}}

    def _remove_relation_between_2_artist(self, artist1, artist2):#{{{
        """Remove a similarity relation"""
        connection = self.get_database_connection()
        connection.execute(
            'DELETE FROM usr_artist_2_artist WHERE artist1 = ? AND artist2 = ?;',
            (artist1, artist2))
        self._clean_artists_table(with_connection=connection)
        self._update_artist(artist_id=artist1, with_connection=connection)
        connection.commit()
        self.close_database_connection(connection)#}}}

    def _remove_artist(self, artist_id, deep=False, with_connection=None):#{{{
        """Remove all artist1 reference"""
        if with_connection:
            connection = with_connection
        else:
            connection = self.get_database_connection()
        if deep:
            connection.execute(
                'DELETE FROM usr_artist_2_artist WHERE artist1 = ? OR artist2 ='
                ' ?;',
                (artist_id, artist_id))
            connection.execute(
                'DELETE FROM artists WHERE id = ?;',
                (artist_id,))
        else:
            connection.execute(
                'DELETE FROM usr_artist_2_artist WHERE artist1 = ?;',
                (artist_id,))
        self._update_artist(artist_id=artist_id, with_connection=connection)
        if not with_connection:
            connection.commit()
            self.close_database_connection(connection)#}}}

    def _insert_artist_match(#{{{
        self, artist1, artist2, match, with_connection=None):
        """write match score to the database.
        Does not update time stamp in table artist/*_updated"""
        if with_connection:
            connection = with_connection
        else:
            connection = self.get_database_connection()
        connection.execute(
            "INSERT INTO usr_artist_2_artist (artist1, artist2, match) VALUES"
            " (?, ?, ?)",
            (artist1, artist2, match))
        if not with_connection:
            connection.commit()
            self.close_database_connection(connection)#}}}

    def _update_artist(self, artist_id, with_connection=None):#{{{
        """write artist information to the database"""
        if with_connection:
            connection = with_connection
        else:
            connection = self.get_database_connection()
        connection.execute(
            "UPDATE artists SET usr_updated = DATETIME('now') WHERE id = ?",
            (artist_id,))
        if not with_connection:
            connection.commit()
            self.close_database_connection(connection)#}}}

    def _update_artist_match(#{{{
        self, artist1, artist2, match, with_connection=None):
        """write match score to the database"""
        if with_connection:
            connection = with_connection
        else:
            connection = self.get_database_connection()
        connection.execute(
            "UPDATE usr_artist_2_artist SET match = ? WHERE artist1 = ? AND"
            " artist2 = ?",
            (match, artist1, artist2))
        if not with_connection:
            connection.commit()
            self.close_database_connection(connection)#}}}

    def _update_similar_artists(self, artist, similar_artists):#{{{
        """write user similar artist information to the database
        """
        # DOC:   similar_artists = list([{'score': match, 'artist': name}])
        #
        artist_id = self.get_artist(artist)[0]
        connection = self.get_database_connection()
        for artist in similar_artists:
            id2 = self.get_artist(
                artist['artist'], with_connection=connection)[0]
            if self._get_artist_match(
                artist_id, id2, with_connection=connection):
                self._update_artist_match(
                    artist_id, id2, artist['score'],
                    with_connection=connection)
                continue
            self._insert_artist_match(
                artist_id, id2, artist['score'],
                with_connection=connection)
        self._update_artist(artist_id, with_connection=connection)
        connection.commit()
        self.close_database_connection(connection)#}}}

    def _clean_artists_table(self, with_connection=None):#{{{
        """Clean orphan artists"""
        if with_connection:
            connection = with_connection
        else:
            connection = self.get_database_connection()
        artists_ids = set([row[0] for row in connection.execute(
            "SELECT id FROM artists")])
        artist_2_artist_ids = set([row[0] for row in connection.execute(
            "SELECT artist1 FROM usr_artist_2_artist")] +
            [row[0] for row in connection.execute(
            "SELECT artist2 FROM usr_artist_2_artist")])
        orphans = artists_ids - artist_2_artist_ids
        while len(orphans) > 0:
            self._remove_artist(orphans.pop(), deep=True,
                    with_connection=connection)
        if not with_connection:
            connection.commit()
            self.close_database_connection(connection)#}}}

    def create_db(self):#{{{
        """ Set up a database for the artist similarity scores
        """
        connection = self.get_database_connection()
        connection.execute(
            'CREATE TABLE IF NOT EXISTS artists (id INTEGER PRIMARY KEY, name'
            ' VARCHAR(100), mbid CHAR(36), lfm_updated DATE, usr_updated DATE)')
        connection.execute(
            'CREATE TABLE IF NOT EXISTS usr_artist_2_artist (artist1 INTEGER,'
            ' artist2 INTEGER, match INTEGER)')
        connection.execute(
            'CREATE TABLE IF NOT EXISTS lfm_artist_2_artist (artist1 INTEGER,'
            ' artist2 INTEGER, match INTEGER)')
        connection.execute(
            'CREATE TABLE IF NOT EXISTS tracks (id INTEGER PRIMARY KEY, artist'
            ' INTEGER, title VARCHAR(100), updated DATE)')
        connection.execute(
            "CREATE INDEX IF NOT EXISTS a2aa1x ON usr_artist_2_artist (artist1)")
        connection.execute(
            "CREATE INDEX IF NOT EXISTS a2aa2x ON usr_artist_2_artist (artist2)")
        connection.execute(
            "CREATE INDEX IF NOT EXISTS lfma2aa1x ON lfm_artist_2_artist (artist1)")
        connection.execute(
            "CREATE INDEX IF NOT EXISTS lfma2aa2x ON lfm_artist_2_artist (artist2)")
        connection.commit()
        self.close_database_connection(connection)#}}}


def main():
    import mpd
    cli = mpd.MPDClient()
    cli.connect(host='libertalia', port='6600')
    cs = cli.currentsong()
    cli.disconnect()
    db = SimaDB(db_path='/home/mxjeff/.local/share/mpd_sima/sima.db')
    #art_u = unicode(cs.get('artist'), 'UTF-8')
    tool =  {'score':int(24),'artist':u'Hint vs Ez3kiel'}
    french = {'score':int(64),'artist':u'Kétanou'}
    #db._update_similar_artists(art_u, [tool, french])
    db._clean_artists_table()
    db.get_artist('Tool', mbid='66fc5bf8-daa4-4241-b378-9bc9077939d2')
    db.get_artist('La Rue Kétanou')
    for a in db.get_artists():
        print a
    #db._set_mbid(artist_id='2', mbid='fddf4c3e-fecc-4fe4-83d7-eb830dcfee2b')
    #for a in db._get_similar_artists_from_db(db.get_artist(u'Led Zeppelin')[0]):
    #    print a
    #for a in db._get_reverse_similar_artists_from_db(int(5)):
    #    print a

# Script starts here
if __name__ == '__main__':
    main()

# VIM MODLINE
# vim: ai ts=4 sw=4 sts=4 expandtab

